# -*- coding: utf-8 -*-
# :Project:   pglast -- Tests ast module
# :Created:   sab 29 mag 2021, 21:25:46
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2021, 2022 Lele Gaifax
#

import pytest

from pglast import ast, enums, parse_sql


def test_compare():
    assert ast.String() != ast.Integer()


def test_bad_values():
    with pytest.raises(ValueError) as e:
        ast.VariableShowStmt({'@': 'SelectStmt'})
    assert "expected 'VariableShowStmt', got 'SelectStmt'" in str(e.value)


def test_call():
    raw = parse_sql('select 1')[0]
    assert raw(0) == {'@': 'RawStmt', 'stmt': ..., 'stmt_len': 0, 'stmt_location': 0}
    assert raw(1)['stmt']['targetList'] == ...
    assert raw(1)['stmt']['targetList'] != 1

    raw = parse_sql('alter table t add constraint c'
                    ' exclude using gist (f with operator(&&))')[0]
    assert raw.stmt.cmds[0].def_(None, skip_none=True)['exclusions'] == (
        ({'@': 'IndexElem',
          'name': 'f',
          'ordering': {'#': 'SortByDir',
                       'name': 'SORTBY_DEFAULT',
                       'value': 0},
          'nulls_ordering': {'#': 'SortByNulls',
                             'name': 'SORTBY_NULLS_DEFAULT',
                             'value': 0}},
         ({'@': 'String', 'sval': '&&'},)),
    )


def test_setattr():
    raw = ast.RawStmt()
    with pytest.raises(ValueError):
        raw.stmt = 'foo'
    raw.stmt = {'@': 'SelectStmt', 'all': True}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt', 'all': 'foo'}
    raw.stmt = {'@': 'SelectStmt',
                'fromClause': ({'@': 'RangeVar',
                                'relname': 'sometable',
                                'relpersistence': 'p'},)}
    raw.stmt = {'@': 'SelectStmt',
                'fromClause': ({'@': 'RangeVar',
                                'relname': 'sometable',
                                'relpersistence': ord('p')},)}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt',
                    'fromClause': ({'@': 'RangeVar',
                                    'relname': 'sometable',
                                    'relpersistence': 'foo'},)}
    raw.stmt = {'@': 'VariableShowStmt', 'name': 'all'}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'VariableShowStmt', 'name': True}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'foo'}}
    raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'LimitOption',
                                                   'name': 'LIMIT_OPTION_DEFAULT'}}
    raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'LimitOption',
                                                   'value': 0}}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'LimitOption'}}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'LimitOption',
                                                       'name': 'foo'}}
    with pytest.raises(ValueError):
        raw.stmt = {'@': 'SelectStmt', 'limitOption': {'#': 'LimitOption',
                                                       'value': -1}}
    raw.stmt = {'@': 'FunctionParameter'}
    raw.stmt.argType = {'@': 'TypeName'}
    raw.stmt = ast.CreateForeignTableStmt()
    raw.stmt.base = {'@': 'CreateStmt'}

def test_issue_97():
    ast.SubLink({
        "@": "SubLink",
        "subLinkType": enums.SubLinkType.ANY_SUBLINK,
        "testexpr": ast.ColumnRef(
            {
                "@": "ColumnRef",
                "fields": (
                    ast.String({"@": "String", "val": "tab"}),
                    ast.String({"@": "String", "val": "_id"}),
                ),
            }
        ),
    })



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from pg_class.h @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017-2022 Lele Gaifax
#

from enum import Enum, IntEnum, IntFlag, auto

try:
    from enum import StrEnum
except ImportError:  # pragma: no cover
    # Python < 3.10
    class StrEnum(str, Enum):
        pass



# #define-ed constants

RELKIND_RELATION = 'r'

RELKIND_INDEX = 'i'

RELKIND_SEQUENCE = 'S'

RELKIND_TOASTVALUE = 't'

RELKIND_VIEW = 'v'

RELKIND_MATVIEW = 'm'

RELKIND_COMPOSITE_TYPE = 'c'

RELKIND_FOREIGN_TABLE = 'f'

RELKIND_PARTITIONED_TABLE = 'p'

RELKIND_PARTITIONED_INDEX = 'I'

RELPERSISTENCE_PERMANENT = 'p'

RELPERSISTENCE_UNLOGGED = 'u'

RELPERSISTENCE_TEMP = 't'

REPLICA_IDENTITY_DEFAULT = 'd'

REPLICA_IDENTITY_NOTHING = 'n'

REPLICA_IDENTITY_FULL = 'f'

REPLICA_IDENTITY_INDEX = 'i'



# -*- coding: utf-8 -*-
# :Project:   pglast -- Wrap PG nodes into a Python AST
# :Created:   sab 27 feb 2021, 19:47:11
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2021, 2022 Lele Gaifax
#

from datetime import date
import json
from keyword import iskeyword
from pathlib import Path
import subprocess
from re import match, sub


CYEARS = f'2021-{date.today().year}'


AST_PY_HEADER = f"""\
# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from struct_defs.json @ %s
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © {CYEARS} Lele Gaifax
#

from collections import namedtuple
from decimal import Decimal
from enum import Enum


SlotTypeInfo = namedtuple('SlotTypeInfo', ['c_type', 'py_type', 'adaptor'])


def _serialize_node(n, depth, ellipsis, skip_none):
    d = {{'@': n.__class__.__name__}}
    for a in n:
        v = _serialize_value(getattr(n, a), depth, ellipsis, skip_none)
        if not skip_none or v is not None:
            d[a] = v
    return d


def _serialize_value(v, depth, ellipsis, skip_none):
    if isinstance(v, Node):
        if depth is None or depth > 0:
            v = _serialize_node(v, None if depth is None else depth - 1,
                                ellipsis, skip_none)
        else:
            v = ellipsis
    elif isinstance(v, tuple):
        if depth is None or depth > 0:
            v = tuple(_serialize_value(i, None if depth is None else depth - 1,
                                       ellipsis, skip_none)
                      for i in v)
        else:
            v = ellipsis
    elif isinstance(v, Enum):
        v = {{'#': v.__class__.__name__, 'name': v.name, 'value': v.value}}
    return v


class Omissis:
    def __eq__(self, other):
        if other is ... or other is self:
            return True
        return False

    def __repr__(self):
        return '…'


Omissis = Omissis()
"Marker value used as default for the ellipsis argument"


class Node:
    "Base class for all AST nodes."

    __slots__ = ('ancestors',)

    def __init__(self, data):
        if not isinstance(data, dict):  # pragma: no cover
            raise ValueError(f'Bad argument, expected a dictionary, got {{type(data)!r}}')
        if '@' not in data:  # pragma: no cover
            raise ValueError('Bad argument, expected a dictionary with a "@" key')
        if data['@'] != self.__class__.__name__:
            raise ValueError(f'Bad argument, wrong "@" value, expected'
                             f' {{self.__class__.__name__!r}}, got {{data["@"]!r}}')

        G = globals()
        for a in self:
            v = data.get(a)
            if v is not None:
                if isinstance(v, dict) and '@' in v:
                    v = G[v['@']](v)
                elif isinstance(v, (tuple, list)):
                    v = tuple(G[i['@']](i) if isinstance(i, dict) and '@' in i else i
                              for i in v)
            setattr(self, a, v)

    def __iter__(self):
        "Iterate over all attribute names of this node."

        return iter(self.__slots__)

    def __repr__(self):
        "Build a representation of the whole node and its subtree, for debug."

        attrs = []
        for a in self:
            if a != 'location':
                v = getattr(self, a)
                if v is not None:
                    attrs.append(f'{{a}}={{v!r}}')
        if attrs:
            attrs = ' ' + ' '.join(attrs)
        else:
            attrs = ''
        return '<' + self.__class__.__name__ + attrs + '>'

    _ATTRS_TO_IGNORE_IN_COMPARISON = {{'location', 'stmt_len', 'stmt_location'}}

    def __eq__(self, other):
        '''
        Compare two nodes, returning ``True`` if they are considered equivalent.

        This is mainly an helper method used by tests: for this reason, two nodes are
        considered equal when all their attributes match, ignoring *positional* ones such as
        ``location``, ``stmt_len`` and ``stmt_location``.
        '''

        if not isinstance(other, type(self)):
            return False
        for a in self:
            if ((a not in self._ATTRS_TO_IGNORE_IN_COMPARISON
                 and getattr(self, a) != getattr(other, a))):
                return False
        return True

    def __call__(self, depth=None, ellipsis=Omissis, skip_none=False):
        '''Serialize the node as a structure made of simple Python data-types.

        :type depth: ``None`` or ``int``
        :param depth: if not ``None``, the maximum depth to reach
        :param ellipsis: the marker value that will be used to replace cut-off branch
        :param bool skip_none: whether ``None``-valued attributes should be elided
        :param bool enum_name: whether Enums will be rendered as their name only
        :return: a :class:`dict` instance

        This performs a top-down recursive visit to the whole AST tree: each :class:`Node`
        instance becomes a dictionary with a special ``@`` key carrying the node type, lists
        becomes tuples and ``Enum`` instances become dictionaries with a special ``#`` key
        carrying the enum name.'''

        return _serialize_node(self, depth, ellipsis, skip_none)

    def __setattr__(self, name, value):
        '''Validate the given `value` and if acceptable assign it to the `name` attribute.

        This tries to coerce the given `value` accordingly with the *ctype* of the
        attribute, raising opportune exception when that is not possible.
        '''

        if value is not None and name in self.__slots__:
            ctype, ptype, adaptor = self.__slots__[name]
            if not isinstance(ptype, tuple):
                ptype = (ptype,)
            if not isinstance(value, ptype):
                raise ValueError(f'Bad value for attribute {{self.__class__.__name__}}'
                                 f'.{{name}}, expected {{ptype}}, got {{type(value)}}:'
                                 f' {{value!r}}')

            if adaptor is not None:
                value = adaptor(value)
            elif ctype != 'char*':
                from pglast import enums

                if hasattr(enums, ctype):
                    enum = getattr(enums, ctype)
                    if not isinstance(value, enum):
                        if isinstance(value, dict) and '#' in value:
                            if value['#'] != ctype:
                                raise ValueError(f'Bad value for attribute'
                                                 f' {{self.__class__.__name__}}.{{name}},'
                                                 f' expected a {{ptype}}, got'
                                                 f' {{value!r}}') from None
                            if 'name' in value:
                                value = value['name']
                            elif 'value' in value:
                                value = value['value']
                            else:
                                raise ValueError(f'Bad value for attribute'
                                                 f' {{self.__class__.__name__}}.{{name}},'
                                                 f' expected a {{ptype}}, got'
                                                 f' {{value!r}}') from None
                        try:
                            if isinstance(value, str) and len(value) > 1:
                                value = enum[value]
                            else:
                                value = enum(value)
                        except (KeyError, ValueError):
                            raise ValueError(f'Bad value for attribute'
                                             f' {{self.__class__.__name__}}.{{name}},'
                                             f' expected a {{ptype}}, got'
                                             f' {{value!r}}') from None
                else:
                    if ctype.endswith('*'):
                        cls = globals().get(ctype[:-1])
                        if cls is None:
                            raise NotImplementedError(f'Unhandled {{ctype!r}} for attribute'
                                                      f' {{self.__class__.__name__}}.{{name}}')
                        if isinstance(value, dict) and '@' in value:
                            value = cls(value)

        super().__setattr__(name, value)


class Expr(Node):
    '''Abstract super class of several *expression* classes.'''

    __slots__ = ()

"""


AST_PYX_HEADER = f"""\
# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from struct_defs.json @ %s
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © {CYEARS} Lele Gaifax
#

#cython: language_level=3

from cpython.ref cimport Py_INCREF
from cpython.tuple cimport PyTuple_New, PyTuple_SET_ITEM

from decimal import Decimal

from pglast import ast, enums
from pglast cimport structs


cdef _pg_bitmapset_to_set(const structs.Bitmapset* bms):
    cdef set result
    cdef int m
    if bms is not NULL:
        result = set()
        m = structs.bms_next_member(bms, -1)
        while m >= 0:
            result.add(m)
            m = structs.bms_next_member(bms, m)
    else:
        result = None
    return result


cdef _pg_str_to_decimal(const char* cstr):
    # Add a trailing zero to truncated floats like "2.", see issues #91 and #100
    cdef s = cstr.decode("utf-8")
    if s.endswith('.'):
        s += '0'
    return ast.Float(Decimal(s))


cdef _pg_list_to_tuple(const structs.List* lst, offset_to_index):
    cdef tuple result
    cdef int i
    if lst is not NULL:
        result = PyTuple_New(lst.length)
        for i in range(lst.length):
            item = create(structs.list_nth(lst, i), offset_to_index)
            Py_INCREF(item)
            PyTuple_SET_ITEM(result, i, item)
    else:
        result = None
    return result
"""


STRUCTS_PXD_HEADER = f"""\
# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from struct_defs.json @ %s
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © {CYEARS} Lele Gaifax
#

#cython: language_level=3

from libc.stdint cimport int16_t, int32_t, uint32_t, uint64_t


cdef extern from "postgres.h":
    ctypedef unsigned char bool

    ctypedef struct Node:
        int type

    ctypedef struct Bitmapset:
        int nwords
        unsigned long *words


cdef extern from "nodes/bitmapset.h":
    ctypedef struct Bitmapset:
        pass

    int bms_next_member(const Bitmapset *a, int prevbit)


cdef extern from "nodes/pg_list.h":
    ctypedef struct List:
        int length

    void* list_nth(List* list, int n)


cdef extern from "nodes/value.h":
    ctypedef struct Integer:
        NodeTag type;
        int ival;

    ctypedef struct Float:
        NodeTag type;
        char *fval

    ctypedef struct Boolean:
        NodeTag type;
        bool boolval;

    ctypedef struct String:
        NodeTag type;
        char *sval;

    ctypedef struct BitString:
        NodeTag type;
        char *bsval;

    int intVal(Integer v)
    double floatVal(Float v)
    bool boolVal(Boolean)
    char* strVal(String v)


# ValUnion is a private type of the A_Const node, we need to redefine it here
ctypedef union ValUnion:
    Node node;
    Integer ival;
    Float fval;
    Boolean boolval;
    String sval;
    BitString bsval;
"""


AST_RST_HEADER = f"""\
.. -*- coding: utf-8 -*-
.. :Project:   pglast -- DO NOT EDIT: generated automatically
.. :Author:    Lele Gaifax <lele@metapensiero.it>
.. :License:   GNU General Public License version 3 or later
.. :Copyright: © {CYEARS} Lele Gaifax
..

.. _pglast.ast:

===================================================================
 :mod:`pglast.ast` --- Python classes representing PG parser nodes
===================================================================

The module implements a set of *data* classes, one for each ``C`` structure defined in several
PostgreSQL headers, primarily those in the `include/nodes/`__ directory.

__ %ssrc/postgres/include/nodes

The :class:`pglast.ast.Node` is an abstract class that implements the common behaviour of all
the concrete classes. In particular any node can be :meth:`compared <pglast.ast.Node.__eq__>`
with another instance, is able to :meth:`serialize <pglast.ast.Node.__call__>` itself and can
be :meth:`altered <pglast.ast.Node.__setattr__>`.

.. module:: pglast.ast

.. autoclass:: Node
   :special-members: __repr__, __eq__, __call__, __setattr__

"""


def get_libpg_query_info():
    "Return a tuple with (version, baseurl) of the libpg_query library."

    version = subprocess.check_output(['git', 'describe', '--all', '--long'],
                                      cwd='libpg_query')
    version = version.decode('utf-8').strip().split('/')[-1]
    remote = subprocess.check_output(['git', 'remote', 'get-url', 'origin'],
                                     cwd='libpg_query')
    remote = remote.decode('utf-8')
    baseurl = '%s/blob/%s/' % (remote[:-5], version[-7:])
    return version, baseurl


def extract_toc(header):
    "Extract the enums and defines with their position in the header."

    toc = {}

    content = header.read_text(encoding='utf-8')

    for lineno, line in enumerate(content.splitlines(), 1):
        if line.startswith(('struct ', 'typedef struct ')):
            m = match(r'(typedef )?struct\s+([\w_]+)', line)
            if m is not None:
                toc[m.group(2)] = lineno

    return toc


def emit_struct_def(name, fields, output):
    output.write(f'    ctypedef struct {name}:\n')

    empty = True
    for field in fields:
        if 'name' not in field or 'c_type' not in field:
            continue

        ctype = field['c_type']
        if ctype in ('Expr', 'Oid'):
            continue
        if name == 'A_Const' and ctype == 'Node':
            ctype = 'ValUnion'
        elif ctype == 'int16':
            ctype = 'int16_t'
        elif ctype in ('bits32', 'int32'):
            ctype = 'int32_t'
        elif ctype == 'uint32':
            ctype = 'uint32_t'
        elif ctype == 'uint64':
            ctype = 'uint64_t'
        if ctype == 'AttrNumber':
            ctype = 'int'
        elif ctype in ('AclMode', 'Index', 'SubTransactionId'):
            ctype = 'unsigned int'
        elif ctype in ('Cardinality', 'Cost'):
            ctype = 'double'
        elif ctype.endswith('*'):
            ctype = f'const {ctype}'

        fname = field['name']
        if iskeyword(fname):
            fname = f'{fname}_ "{fname}"'

        output.write(f'        {ctype} {fname}\n')
        empty = False

    if empty:
        output.write('        pass\n')


def emit_generic_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = data.{name}
''')


def emit_bool_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = bool(data.{name})
''')


def emit_valunion_attr(name, ctype, output):
    # Must inline this code, because ValUnion is a private type of the A_Const node,
    # so it's difficult to cleanly pass the value to a subfunction
    output.write(f'''\
    cdef object v_{name}
    if data.isnull:
        v_{name} = None
    elif data.{name}.boolval.type == structs.T_Boolean:
        v_{name} = ast.Boolean(data.{name}.boolval.boolval)
    elif data.{name}.ival.type == structs.T_Integer:
        v_{name} = ast.Integer(data.{name}.ival.ival)
    elif data.{name}.fval.type == structs.T_Float:
        v_{name} = _pg_str_to_decimal(data.{name}.fval.fval)
    elif data.{name}.bsval.type == structs.T_BitString:
        v_{name} = ast.BitString(data.{name}.bsval.bsval.decode("utf-8"))
    elif data.{name}.sval.type == structs.T_String:
        v_{name} = ast.String(data.{name}.sval.sval.decode("utf-8"))
    else:
        v_{name} = data.{name}.node
''')


def emit_str_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = data.{name}.decode("utf-8") if data.{name} is not NULL else None
''')


def emit_char_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = chr(data.{name})
''')


def emit_list_attr(name, ctype, output):
    output.write(f'''\
    cdef tuple v_{name} = _pg_list_to_tuple(data.{name}, offset_to_index)
''')


def emit_node_attr(name, ctype, output):
    output.write(f'''\
    cdef v_{name} = create(&data.{name}, offset_to_index)
''')


def emit_create_stmt_attr(name, ctype, output):
    output.write(f'''
    cdef object v_{name} = create_CreateStmt(<structs.CreateStmt*> data, offset_to_index)
''')


def emit_nodeptr_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = create(data.{name}, offset_to_index) if data.{name} is not NULL else None
''')


def emit_no_attr(name, ctype, output):
    # output.write(f'#        cdef object v_{name} = data.{name}\n')
    pass


def emit_int_enum_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = getattr(enums, {ctype!r})(data.{name})
''')


def emit_str_enum_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = getattr(enums, {ctype!r})(chr(data.{name}))
''')


def emit_bitmapset_attr(name, ctype, output):
    output.write(f'''\
    cdef set v_{name} = _pg_bitmapset_to_set(data.{name})
''')


def emit_location_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = offset_to_index(data.{name})
''')


def emit_stmt_len_attr(name, ctype, output):
    output.write(f'''\
    cdef object v_{name} = offset_to_index(data.stmt_location + data.{name}) - offset_to_index(data.stmt_location)
''')


def emitter_for(fname, ctype, enums):
    from pglast import enums as eimpl

    if fname == 'location' and ctype == 'int' or fname == 'stmt_location':
        emitter = emit_location_attr
    elif fname == 'stmt_len':
        emitter = emit_stmt_len_attr
    elif ctype == 'List*':
        emitter = emit_list_attr
    elif ctype == 'CreateStmt':
        emitter = emit_create_stmt_attr
    elif ctype == 'Expr':
        emitter = emit_no_attr
    elif ctype == 'NodeTag':
        emitter = emit_no_attr
    elif ctype == 'ValUnion':
        emitter = emit_valunion_attr
    elif ctype == 'char*':
        emitter = emit_str_attr
    elif ctype == 'char':
        emitter = emit_char_attr
    elif ctype == 'bool':
        emitter = emit_bool_attr
    elif ctype == 'Bitmapset*':
        emitter = emit_bitmapset_attr
    elif ctype.endswith('*'):
        emitter = emit_nodeptr_attr
    elif ctype in enums:
        if issubclass(getattr(eimpl, ctype), eimpl.IntEnum):
            emitter = emit_int_enum_attr
        else:
            emitter = emit_str_enum_attr
    else:
        emitter = emit_generic_attr

    return emitter


def emit_node_def(name, fields, enums, url, output, doc):
    attrs = []
    attrs_to_ignore_in_comparison = set()

    superclass = 'Node'

    for field in fields:
        if 'name' not in field or 'c_type' not in field:
            continue

        ctype = field['c_type']

        if ctype == 'Oid':
            continue

        fname = field['name']
        if iskeyword(fname):
            fname = f'{fname}_'

        if ctype == 'CoercionForm':
            attrs_to_ignore_in_comparison.add(fname)

        emitter = emitter_for(fname, ctype, enums)
        if ctype == 'Expr':
            superclass = 'Expr'

        comment = field.get('comment')
        if comment:
            comment = comment.strip()
            if comment.startswith('/*'):
                comment = comment[2:-2].strip()
            comment = sub(r'\t+( \* (?![A-Z] = ))?', ' ', comment)
            comment = sub(r'\*-+\s*', '', comment)
            comment = sub(r'-+\n', '', comment)
            comment = sub(r'\n +', '\n      ', comment)
            comment = sub(r'\*\)', '\\*)', comment)
            comment = comment.strip()
            comment = comment[0].upper() + comment[1:]
            if comment.lower() == 'see above':
                comment = ''

        attrs.append((fname, ctype, comment, emitter))

    real_attrs = []
    if attrs:
        for attr, ctype, comment, emitter in attrs:
            if emitter is emit_no_attr:
                continue
            real_attrs.append((attr, ctype))

    output.write(f'''\


class {name}({superclass}):
    __slots__ = {{{', '.join(repr(a)+': '+repr(t) for a, t in real_attrs)}}}  # noqa: E501

''')

    if attrs_to_ignore_in_comparison:
        output.write(f'''\
    _ATTRS_TO_IGNORE_IN_COMPARISON = {superclass}._ATTRS_TO_IGNORE_IN_COMPARISON | {repr(attrs_to_ignore_in_comparison)}

''')
    if real_attrs:
        output.write(f'''\
    def __init__(self, {', '.join(f'{attr}=None' for attr, __ in real_attrs)}):  # pragma: no cover  # noqa: E501
''')

        if len(real_attrs) > 1:
            output.write(f'''\
        if (({real_attrs[0][0]} is not None
             and {' is '.join(attr for attr, __ in real_attrs[1:])} is None  # noqa: E501
             and isinstance({real_attrs[0][0]}, dict)
             and '@' in {real_attrs[0][0]})):
            super().__init__({real_attrs[0][0]})
        else:
''')
            for a, v in real_attrs:
                output.write(f'            self.{a} = {a}\n')
        else:
            output.write(f'''
        if (({real_attrs[0][0]} is not None
             and isinstance({real_attrs[0][0]}, dict)
             and '@' in {real_attrs[0][0]})):
            super().__init__({real_attrs[0][0]})
        else:
''')
            for a, v in real_attrs:
                output.write(f'            self.{a} = {a}\n')
    else:
        output.write('''\
    def __init__(self):  # pragma: no cover
        pass
''')

    doc.write(f'''
.. class:: {name}({', '.join(f'{attr}=None' for attr, __ in real_attrs)})

   Wrapper for the `homonymous <{url}>`__ parser node.

''')
    for attr, ctype, comment, emitter in attrs:
        if emitter is emit_no_attr:
            continue
        if ctype == 'List*':
            type = 'tuple'
        elif ctype in ('char', 'char*'):
            type = 'str'
        elif ctype == 'Node*':
            type = 'Node'
        elif ctype in enums:
            type = f'{ctype}'
        else:
            type = ctype
        doc.write(f'   .. attribute:: {attr}\n')
        doc.write(f'      :type: {type}\n\n')
        if comment:
            doc.write(f'      {comment}\n\n')


def emit_node_create_function(nodes, enums, output):
    from pglast import enums as eimpl

    nnames = set(n[0] for n in nodes)

    for name, fields in nodes:
        attrs = []
        real_attrs = []
        for field in fields:
            if 'name' not in field or 'c_type' not in field:
                continue

            ctype = field['c_type']
            if ctype in ('Expr', 'Oid'):
                continue

            fname = field['name']
            if iskeyword(fname):
                fname = f'{fname}_'

            emitter = emitter_for(fname, ctype, enums)

            attrs.append((fname, ctype, emitter))
            if emitter is not emit_no_attr:
                real_attrs.append((fname, ctype))

        output.write(f'''\


cdef create_{name}(structs.{name}* data, offset_to_index):
''')
        for attr, ctype, emitter in attrs:
            emitter(attr, ctype, output)
        output.write(f'''\
    return ast.{name}({', '.join(f'v_{attr}' for attr, __ in real_attrs)})
''')

    output.write('''\


cdef create(void* data, offset_to_index):
    if data is NULL:
        return None

    cdef tuple t
    cdef int i
    cdef str s
    cdef int tag = structs.nodeTag(data)

''')

    first = True
    for tag in eimpl.NodeTag:
        name = tag.name[2:]
        if name in nnames:
            output.write('    ')
            output.write('if' if first else 'elif')
            output.write(f' tag == structs.{tag.name}:\n')
            output.write(f'        return create_{name}(<structs.{name}*> data, offset_to_index)\n')
            first = False
        elif name == 'List':
            output.write('''\
    elif tag == structs.T_List:
        return _pg_list_to_tuple(<structs.List *> data, offset_to_index)
''')

    output.write('''\
    raise ValueError("Unhandled tag: %s" % tag)
''')


def emit_valunion_def(output):
    output.write("""

class ValUnion(Node):
   '''Represent `ValUnion`__ value.

   __ https://git.postgresql.org/gitweb/?p=postgresql.git;a=blob;
      f=src/include/nodes/parsenodes.h;hb=c5dc80c1bc216f0e21a2f79f5e0415c2d4cfb35d#l309
   '''

   __slots__ = {'val': SlotTypeInfo('ValUnion', Node, None)}

   def __init__(self, value=None):  # pragma: no cover  # noqa: E501
       if ((value is not None
            and isinstance(value, dict)
            and '@' in value)):
           super().__init__(value)
       else:
           self.val = value
""")


def workhorse(args):
    libpg_query_version, libpg_query_baseurl = get_libpg_query_info()

    pgq_dir = Path('libpg_query')
    pg_inc_dir = pgq_dir / 'src' / 'postgres' / 'include'

    with (pgq_dir / 'srcdata' / 'struct_defs.json').open(encoding='utf-8') as f:
        structs = json.load(f)

    ctypes = set()
    for header in ('nodes/parsenodes', 'nodes/primnodes'):
        for name in structs[header]:
            fields = structs[header][name]['fields']
            for field in fields:
                if 'c_type' in field:
                    ctypes.add(field['c_type'])

    with (pgq_dir / 'srcdata' / 'all_known_enums.json').open(encoding='utf-8') as f:
        enums = sorted(json.load(f))

    with (pgq_dir / 'srcdata' / 'enum_defs.json').open(encoding='utf-8') as f:
        node_tags = [e['name']
                     for e in json.load(f)['nodes/nodes']['NodeTag']['values']
                     if 'name' in e]

    linenos = {}

    structs_pxd = args.output_dir / 'structs.pxd'
    with structs_pxd.open('w', encoding='utf-8') as output:
        output.write(STRUCTS_PXD_HEADER % libpg_query_version)

        output.write('\n\ncdef extern from *:')
        for name in enums:
            output.write(f'\n    ctypedef enum {name}:\n')
            if name == 'NodeTag':
                for tag in node_tags:
                    output.write(f'        {tag}\n')
            else:
                output.write('        pass\n')
        output.write('\n    int nodeTag(void* data)\n')

        nodes = []
        for header in ('nodes/parsenodes', 'nodes/primnodes', 'nodes/value'):
            output.write(f'\n\ncdef extern from "{header}.h":\n')
            toc = extract_toc(pg_inc_dir / (header + '.h'))
            for name in toc:
                linenos[name] = (header, toc[name])
            defs = structs[header]
            first = True
            for name in defs:
                fields = defs[name]['fields']
                if name not in ('Const', 'NextValueExpr'):
                    if name != 'Expr':
                        # Omit the Expr node, because it is hand written in the
                        # ast.py header above: also, it is an abstract class,
                        # never instantiated directly
                        nodes.append((name, fields))
                    if first:
                        first = False
                    else:
                        output.write('\n')
                    emit_struct_def(name, fields, output)

    ast_py = args.output_dir / 'ast.py'
    with ast_py.open('w', encoding='utf-8') as output, \
         args.rstdoc.open('w', encoding='utf-8') as doc:
        output.write(AST_PY_HEADER % libpg_query_version)

        doc.write(AST_RST_HEADER % libpg_query_baseurl)

        for name, fields in sorted(nodes):
            header, lineno = linenos[name]
            url = f'{libpg_query_baseurl}src/postgres/include/{header}.h#L{lineno}'
            if name == 'A_Const':
                emit_valunion_def(output)
                for f in fields:
                    if f['name'] == 'val':
                        f['c_type'] = 'ValUnion'
                        break
            emit_node_def(name, fields, enums, url, output, doc)

        output.write('''

def _fixup_attribute_types_in_slots():
    G = globals()

    def traverse_sub_classes(cls):
        for subc in cls.__subclasses__():
            yield subc
            yield from traverse_sub_classes(subc)

    for cls in traverse_sub_classes(Node):
        slots = cls.__slots__
        if not (slots
                and isinstance(slots, dict)
                and isinstance(next(iter(slots.values())), str)):
            continue
        for attr in slots:
            adaptor = None
            ctype = slots[attr]
            if ctype == 'List*':
                ptype = (list, tuple)

                def adaptor(value):
                    return tuple(G[i['@']](i)
                                 if isinstance(i, dict) and '@' in i
                                 else i
                                 for i in value)
            elif ctype == 'bool':
                ptype = (bool, int)
                adaptor = bool
            elif ctype == 'char':
                ptype = (str, int)

                def adaptor(value):
                    if isinstance(value, int):
                        value = chr(value)
                    elif len(value) != 1:
                        raise ValueError(f'Bad value for attribute {cls.__name__}.{attr},'
                                         f' expected a single char, got {value!r}')
                    return value
            elif cls is Float and ctype == 'char*':
                ptype = (str, float, Decimal)
                adaptor = Decimal
            elif ctype == 'char*':
                ptype = str
            elif ctype in ('Expr*', 'Node*'):
                ptype = (dict, list, tuple, Node)

                def adaptor(value):
                    if isinstance(value, dict):
                        if '@' in value:
                            value = G[value['@']](value)
                    elif isinstance(value, (list, tuple)):
                        value = tuple(G[i['@']](i)
                                      if isinstance(i, dict) and '@' in i
                                      else i
                                      for i in value)
                    return value
            elif ctype in ('int', 'int16', 'bits32', 'int32', 'uint32', 'uint64',
                           'AttrNumber', 'AclMode', 'Index', 'SubTransactionId'):
                ptype = int
            elif ctype in ('Cardinality', 'Cost'):
                ptype = float
            elif ctype == 'CreateStmt':
                ptype = (dict, CreateStmt)

                def adaptor(value):
                    if isinstance(value, dict):
                        if '@' in value:
                            value = G[value['@']](value)
                    return value
            elif ctype == 'Bitmapset*':
                ptype = (list, set, tuple)

                def adaptor(value):
                    if isinstance(value, (list, tuple)):
                        return set(value)
                    else:
                        return value
            elif ctype == 'ValUnion':
                ptype = Node
            else:
                from pglast import enums

                if hasattr(enums, ctype):
                    ptype = (int, str, dict, getattr(enums, ctype))
                else:
                    if ctype.endswith('*'):
                        ptype = G.get(ctype[:-1])
                        if ptype is None:
                            raise NotImplementedError(f'unknown {ctype!r}') from None
                        else:
                            ptype = (dict, ptype)
            slots[attr] = SlotTypeInfo(ctype, ptype, adaptor)


_fixup_attribute_types_in_slots()
del _fixup_attribute_types_in_slots
''')

    ast_pyx = args.output_dir / 'ast.pyx'
    with ast_pyx.open('w', encoding='utf-8') as output:
        output.write(AST_PYX_HEADER % libpg_query_version)
        emit_node_create_function(nodes, enums, output)


def main():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(description="PG structs extractor",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('output_dir', type=Path,
                        help="where Cython sources will be created")
    parser.add_argument('rstdoc', type=Path,
                        help="reST documentation to be created")

    args = parser.parse_args()

    workhorse(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   pglast -- DO NOT EDIT: automatically extracted from struct_defs.json @ 15-4.0.0-0-g1ada550
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2021-2022 Lele Gaifax
#

#cython: language_level=3

from libc.stdint cimport int16_t, int32_t, uint32_t, uint64_t


cdef extern from "postgres.h":
    ctypedef unsigned char bool

    ctypedef struct Node:
        int type

    ctypedef struct Bitmapset:
        int nwords
        unsigned long *words


cdef extern from "nodes/bitmapset.h":
    ctypedef struct Bitmapset:
        pass

    int bms_next_member(const Bitmapset *a, int prevbit)


cdef extern from "nodes/pg_list.h":
    ctypedef struct List:
        int length

    void* list_nth(List* list, int n)


cdef extern from "nodes/value.h":
    ctypedef struct Integer:
        NodeTag type;
        int ival;

    ctypedef struct Float:
        NodeTag type;
        char *fval

    ctypedef struct Boolean:
        NodeTag type;
        bool boolval;

    ctypedef struct String:
        NodeTag type;
        char *sval;

    ctypedef struct BitString:
        NodeTag type;
        char *bsval;

    int intVal(Integer v)
    double floatVal(Float v)
    bool boolVal(Boolean)
    char* strVal(String v)


# ValUnion is a private type of the A_Const node, we need to redefine it here
ctypedef union ValUnion:
    Node node;
    Integer ival;
    Float fval;
    Boolean boolval;
    String sval;
    BitString bsval;


cdef extern from *:
    ctypedef enum A_Expr_Kind:
        pass

    ctypedef enum AggSplit:
        pass

    ctypedef enum AggStrategy:
        pass

    ctypedef enum AlterPublicationAction:
        pass

    ctypedef enum AlterSubscriptionType:
        pass

    ctypedef enum AlterTSConfigType:
        pass

    ctypedef enum AlterTableType:
        pass

    ctypedef enum BoolExprType:
        pass

    ctypedef enum BoolTestType:
        pass

    ctypedef enum CTEMaterialize:
        pass

    ctypedef enum CmdType:
        pass

    ctypedef enum CoercionContext:
        pass

    ctypedef enum CoercionForm:
        pass

    ctypedef enum ConstrType:
        pass

    ctypedef enum DefElemAction:
        pass

    ctypedef enum DiscardMode:
        pass

    ctypedef enum DropBehavior:
        pass

    ctypedef enum FetchDirection:
        pass

    ctypedef enum FunctionParameterMode:
        pass

    ctypedef enum GrantTargetType:
        pass

    ctypedef enum GroupingSetKind:
        pass

    ctypedef enum ImportForeignSchemaType:
        pass

    ctypedef enum JoinType:
        pass

    ctypedef enum LimitOption:
        pass

    ctypedef enum LockClauseStrength:
        pass

    ctypedef enum LockTupleMode:
        pass

    ctypedef enum LockWaitPolicy:
        pass

    ctypedef enum MinMaxOp:
        pass

    ctypedef enum NodeTag:
        T_Invalid
        T_IndexInfo
        T_ExprContext
        T_ProjectionInfo
        T_JunkFilter
        T_OnConflictSetState
        T_MergeActionState
        T_ResultRelInfo
        T_EState
        T_TupleTableSlot
        T_Plan
        T_Result
        T_ProjectSet
        T_ModifyTable
        T_Append
        T_MergeAppend
        T_RecursiveUnion
        T_BitmapAnd
        T_BitmapOr
        T_Scan
        T_SeqScan
        T_SampleScan
        T_IndexScan
        T_IndexOnlyScan
        T_BitmapIndexScan
        T_BitmapHeapScan
        T_TidScan
        T_TidRangeScan
        T_SubqueryScan
        T_FunctionScan
        T_ValuesScan
        T_TableFuncScan
        T_CteScan
        T_NamedTuplestoreScan
        T_WorkTableScan
        T_ForeignScan
        T_CustomScan
        T_Join
        T_NestLoop
        T_MergeJoin
        T_HashJoin
        T_Material
        T_Memoize
        T_Sort
        T_IncrementalSort
        T_Group
        T_Agg
        T_WindowAgg
        T_Unique
        T_Gather
        T_GatherMerge
        T_Hash
        T_SetOp
        T_LockRows
        T_Limit
        T_NestLoopParam
        T_PlanRowMark
        T_PartitionPruneInfo
        T_PartitionedRelPruneInfo
        T_PartitionPruneStepOp
        T_PartitionPruneStepCombine
        T_PlanInvalItem
        T_PlanState
        T_ResultState
        T_ProjectSetState
        T_ModifyTableState
        T_AppendState
        T_MergeAppendState
        T_RecursiveUnionState
        T_BitmapAndState
        T_BitmapOrState
        T_ScanState
        T_SeqScanState
        T_SampleScanState
        T_IndexScanState
        T_IndexOnlyScanState
        T_BitmapIndexScanState
        T_BitmapHeapScanState
        T_TidScanState
        T_TidRangeScanState
        T_SubqueryScanState
        T_FunctionScanState
        T_TableFuncScanState
        T_ValuesScanState
        T_CteScanState
        T_NamedTuplestoreScanState
        T_WorkTableScanState
        T_ForeignScanState
        T_CustomScanState
        T_JoinState
        T_NestLoopState
        T_MergeJoinState
        T_HashJoinState
        T_MaterialState
        T_MemoizeState
        T_SortState
        T_IncrementalSortState
        T_GroupState
        T_AggState
        T_WindowAggState
        T_UniqueState
        T_GatherState
        T_GatherMergeState
        T_HashState
        T_SetOpState
        T_LockRowsState
        T_LimitState
        T_Alias
        T_RangeVar
        T_TableFunc
        T_Var
        T_Const
        T_Param
        T_Aggref
        T_GroupingFunc
        T_WindowFunc
        T_SubscriptingRef
        T_FuncExpr
        T_NamedArgExpr
        T_OpExpr
        T_DistinctExpr
        T_NullIfExpr
        T_ScalarArrayOpExpr
        T_BoolExpr
        T_SubLink
        T_SubPlan
        T_AlternativeSubPlan
        T_FieldSelect
        T_FieldStore
        T_RelabelType
        T_CoerceViaIO
        T_ArrayCoerceExpr
        T_ConvertRowtypeExpr
        T_CollateExpr
        T_CaseExpr
        T_CaseWhen
        T_CaseTestExpr
        T_ArrayExpr
        T_RowExpr
        T_RowCompareExpr
        T_CoalesceExpr
        T_MinMaxExpr
        T_SQLValueFunction
        T_XmlExpr
        T_NullTest
        T_BooleanTest
        T_CoerceToDomain
        T_CoerceToDomainValue
        T_SetToDefault
        T_CurrentOfExpr
        T_NextValueExpr
        T_InferenceElem
        T_TargetEntry
        T_RangeTblRef
        T_JoinExpr
        T_FromExpr
        T_OnConflictExpr
        T_IntoClause
        T_ExprState
        T_WindowFuncExprState
        T_SetExprState
        T_SubPlanState
        T_DomainConstraintState
        T_PlannerInfo
        T_PlannerGlobal
        T_RelOptInfo
        T_IndexOptInfo
        T_ForeignKeyOptInfo
        T_ParamPathInfo
        T_Path
        T_IndexPath
        T_BitmapHeapPath
        T_BitmapAndPath
        T_BitmapOrPath
        T_TidPath
        T_TidRangePath
        T_SubqueryScanPath
        T_ForeignPath
        T_CustomPath
        T_NestPath
        T_MergePath
        T_HashPath
        T_AppendPath
        T_MergeAppendPath
        T_GroupResultPath
        T_MaterialPath
        T_MemoizePath
        T_UniquePath
        T_GatherPath
        T_GatherMergePath
        T_ProjectionPath
        T_ProjectSetPath
        T_SortPath
        T_IncrementalSortPath
        T_GroupPath
        T_UpperUniquePath
        T_AggPath
        T_GroupingSetsPath
        T_MinMaxAggPath
        T_WindowAggPath
        T_SetOpPath
        T_RecursiveUnionPath
        T_LockRowsPath
        T_ModifyTablePath
        T_LimitPath
        T_EquivalenceClass
        T_EquivalenceMember
        T_PathKey
        T_PathKeyInfo
        T_PathTarget
        T_RestrictInfo
        T_IndexClause
        T_PlaceHolderVar
        T_SpecialJoinInfo
        T_AppendRelInfo
        T_RowIdentityVarInfo
        T_PlaceHolderInfo
        T_MinMaxAggInfo
        T_PlannerParamItem
        T_RollupData
        T_GroupingSetData
        T_StatisticExtInfo
        T_MergeAction
        T_AllocSetContext
        T_SlabContext
        T_GenerationContext
        T_Integer
        T_Float
        T_Boolean
        T_String
        T_BitString
        T_List
        T_IntList
        T_OidList
        T_ExtensibleNode
        T_RawStmt
        T_Query
        T_PlannedStmt
        T_InsertStmt
        T_DeleteStmt
        T_UpdateStmt
        T_MergeStmt
        T_SelectStmt
        T_ReturnStmt
        T_PLAssignStmt
        T_AlterTableStmt
        T_AlterTableCmd
        T_AlterDomainStmt
        T_SetOperationStmt
        T_GrantStmt
        T_GrantRoleStmt
        T_AlterDefaultPrivilegesStmt
        T_ClosePortalStmt
        T_ClusterStmt
        T_CopyStmt
        T_CreateStmt
        T_DefineStmt
        T_DropStmt
        T_TruncateStmt
        T_CommentStmt
        T_FetchStmt
        T_IndexStmt
        T_CreateFunctionStmt
        T_AlterFunctionStmt
        T_DoStmt
        T_RenameStmt
        T_RuleStmt
        T_NotifyStmt
        T_ListenStmt
        T_UnlistenStmt
        T_TransactionStmt
        T_ViewStmt
        T_LoadStmt
        T_CreateDomainStmt
        T_CreatedbStmt
        T_DropdbStmt
        T_VacuumStmt
        T_ExplainStmt
        T_CreateTableAsStmt
        T_CreateSeqStmt
        T_AlterSeqStmt
        T_VariableSetStmt
        T_VariableShowStmt
        T_DiscardStmt
        T_CreateTrigStmt
        T_CreatePLangStmt
        T_CreateRoleStmt
        T_AlterRoleStmt
        T_DropRoleStmt
        T_LockStmt
        T_ConstraintsSetStmt
        T_ReindexStmt
        T_CheckPointStmt
        T_CreateSchemaStmt
        T_AlterDatabaseStmt
        T_AlterDatabaseRefreshCollStmt
        T_AlterDatabaseSetStmt
        T_AlterRoleSetStmt
        T_CreateConversionStmt
        T_CreateCastStmt
        T_CreateOpClassStmt
        T_CreateOpFamilyStmt
        T_AlterOpFamilyStmt
        T_PrepareStmt
        T_ExecuteStmt
        T_DeallocateStmt
        T_DeclareCursorStmt
        T_CreateTableSpaceStmt
        T_DropTableSpaceStmt
        T_AlterObjectDependsStmt
        T_AlterObjectSchemaStmt
        T_AlterOwnerStmt
        T_AlterOperatorStmt
        T_AlterTypeStmt
        T_DropOwnedStmt
        T_ReassignOwnedStmt
        T_CompositeTypeStmt
        T_CreateEnumStmt
        T_CreateRangeStmt
        T_AlterEnumStmt
        T_AlterTSDictionaryStmt
        T_AlterTSConfigurationStmt
        T_CreateFdwStmt
        T_AlterFdwStmt
        T_CreateForeignServerStmt
        T_AlterForeignServerStmt
        T_CreateUserMappingStmt
        T_AlterUserMappingStmt
        T_DropUserMappingStmt
        T_AlterTableSpaceOptionsStmt
        T_AlterTableMoveAllStmt
        T_SecLabelStmt
        T_CreateForeignTableStmt
        T_ImportForeignSchemaStmt
        T_CreateExtensionStmt
        T_AlterExtensionStmt
        T_AlterExtensionContentsStmt
        T_CreateEventTrigStmt
        T_AlterEventTrigStmt
        T_RefreshMatViewStmt
        T_ReplicaIdentityStmt
        T_AlterSystemStmt
        T_CreatePolicyStmt
        T_AlterPolicyStmt
        T_CreateTransformStmt
        T_CreateAmStmt
        T_CreatePublicationStmt
        T_AlterPublicationStmt
        T_CreateSubscriptionStmt
        T_AlterSubscriptionStmt
        T_DropSubscriptionStmt
        T_CreateStatsStmt
        T_AlterCollationStmt
        T_CallStmt
        T_AlterStatsStmt
        T_A_Expr
        T_ColumnRef
        T_ParamRef
        T_A_Const
        T_FuncCall
        T_A_Star
        T_A_Indices
        T_A_Indirection
        T_A_ArrayExpr
        T_ResTarget
        T_MultiAssignRef
        T_TypeCast
        T_CollateClause
        T_SortBy
        T_WindowDef
        T_RangeSubselect
        T_RangeFunction
        T_RangeTableSample
        T_RangeTableFunc
        T_RangeTableFuncCol
        T_TypeName
        T_ColumnDef
        T_IndexElem
        T_StatsElem
        T_Constraint
        T_DefElem
        T_RangeTblEntry
        T_RangeTblFunction
        T_TableSampleClause
        T_WithCheckOption
        T_SortGroupClause
        T_GroupingSet
        T_WindowClause
        T_ObjectWithArgs
        T_AccessPriv
        T_CreateOpClassItem
        T_TableLikeClause
        T_FunctionParameter
        T_LockingClause
        T_RowMarkClause
        T_XmlSerialize
        T_WithClause
        T_InferClause
        T_OnConflictClause
        T_CTESearchClause
        T_CTECycleClause
        T_CommonTableExpr
        T_MergeWhenClause
        T_RoleSpec
        T_TriggerTransition
        T_PartitionElem
        T_PartitionSpec
        T_PartitionBoundSpec
        T_PartitionRangeDatum
        T_PartitionCmd
        T_VacuumRelation
        T_PublicationObjSpec
        T_PublicationTable
        T_IdentifySystemCmd
        T_BaseBackupCmd
        T_CreateReplicationSlotCmd
        T_DropReplicationSlotCmd
        T_ReadReplicationSlotCmd
        T_StartReplicationCmd
        T_TimeLineHistoryCmd
        T_TriggerData
        T_EventTriggerData
        T_ReturnSetInfo
        T_WindowObjectData
        T_TIDBitmap
        T_InlineCodeBlock
        T_FdwRoutine
        T_IndexAmRoutine
        T_TableAmRoutine
        T_TsmRoutine
        T_ForeignKeyCacheInfo
        T_CallContext
        T_SupportRequestSimplify
        T_SupportRequestSelectivity
        T_SupportRequestCost
        T_SupportRequestRows
        T_SupportRequestIndexCondition
        T_SupportRequestWFuncMonotonic

    ctypedef enum NullTestType:
        pass

    ctypedef enum ObjectType:
        pass

    ctypedef enum OnCommitAction:
        pass

    ctypedef enum OnConflictAction:
        pass

    ctypedef enum OverridingKind:
        pass

    ctypedef enum ParamKind:
        pass

    ctypedef enum PartitionRangeDatumKind:
        pass

    ctypedef enum PublicationObjSpecType:
        pass

    ctypedef enum QuerySource:
        pass

    ctypedef enum RTEKind:
        pass

    ctypedef enum ReindexObjectType:
        pass

    ctypedef enum RoleSpecType:
        pass

    ctypedef enum RoleStmtType:
        pass

    ctypedef enum RowCompareType:
        pass

    ctypedef enum SQLValueFunctionOp:
        pass

    ctypedef enum ScanDirection:
        pass

    ctypedef enum SetOpCmd:
        pass

    ctypedef enum SetOpStrategy:
        pass

    ctypedef enum SetOperation:
        pass

    ctypedef enum SetQuantifier:
        pass

    ctypedef enum SortByDir:
        pass

    ctypedef enum SortByNulls:
        pass

    ctypedef enum SubLinkType:
        pass

    ctypedef enum TableLikeOption:
        pass

    ctypedef enum TransactionStmtKind:
        pass

    ctypedef enum VacOptValue:
        pass

    ctypedef enum VariableSetKind:
        pass

    ctypedef enum ViewCheckOption:
        pass

    ctypedef enum WCOKind:
        pass

    ctypedef enum XmlExprOp:
        pass

    ctypedef enum XmlOptionType:
        pass

    ctypedef enum pg_enc:
        pass

    ctypedef enum vartag_external:
        pass

    ctypedef enum yytokentype:
        pass

    int nodeTag(void* data)


cdef extern from "nodes/parsenodes.h":
    ctypedef struct Query:
        NodeTag type
        CmdType commandType
        QuerySource querySource
        uint64_t queryId
        bool canSetTag
        const Node* utilityStmt
        int resultRelation
        bool hasAggs
        bool hasWindowFuncs
        bool hasTargetSRFs
        bool hasSubLinks
        bool hasDistinctOn
        bool hasRecursive
        bool hasModifyingCTE
        bool hasForUpdate
        bool hasRowSecurity
        bool isReturn
        const List* cteList
        const List* rtable
        const FromExpr* jointree
        const List* mergeActionList
        bool mergeUseOuterJoin
        const List* targetList
        OverridingKind override
        const OnConflictExpr* onConflict
        const List* returningList
        const List* groupClause
        bool groupDistinct
        const List* groupingSets
        const Node* havingQual
        const List* windowClause
        const List* distinctClause
        const List* sortClause
        const Node* limitOffset
        const Node* limitCount
        LimitOption limitOption
        const List* rowMarks
        const Node* setOperations
        const List* constraintDeps
        const List* withCheckOptions
        int stmt_location
        int stmt_len

    ctypedef struct TypeName:
        NodeTag type
        const List* names
        bool setof
        bool pct_type
        const List* typmods
        int32_t typemod
        const List* arrayBounds
        int location

    ctypedef struct ColumnRef:
        NodeTag type
        const List* fields
        int location

    ctypedef struct ParamRef:
        NodeTag type
        int number
        int location

    ctypedef struct A_Expr:
        NodeTag type
        A_Expr_Kind kind
        const List* name
        const Node* lexpr
        const Node* rexpr
        int location

    ctypedef struct TypeCast:
        NodeTag type
        const Node* arg
        const TypeName* typeName
        int location

    ctypedef struct CollateClause:
        NodeTag type
        const Node* arg
        const List* collname
        int location

    ctypedef struct RoleSpec:
        NodeTag type
        RoleSpecType roletype
        const char* rolename
        int location

    ctypedef struct FuncCall:
        NodeTag type
        const List* funcname
        const List* args
        const List* agg_order
        const Node* agg_filter
        const WindowDef* over
        bool agg_within_group
        bool agg_star
        bool agg_distinct
        bool func_variadic
        CoercionForm funcformat
        int location

    ctypedef struct A_Star:
        NodeTag type

    ctypedef struct A_Indices:
        NodeTag type
        bool is_slice
        const Node* lidx
        const Node* uidx

    ctypedef struct A_Indirection:
        NodeTag type
        const Node* arg
        const List* indirection

    ctypedef struct A_ArrayExpr:
        NodeTag type
        const List* elements
        int location

    ctypedef struct ResTarget:
        NodeTag type
        const char* name
        const List* indirection
        const Node* val
        int location

    ctypedef struct MultiAssignRef:
        NodeTag type
        const Node* source
        int colno
        int ncolumns

    ctypedef struct SortBy:
        NodeTag type
        const Node* node
        SortByDir sortby_dir
        SortByNulls sortby_nulls
        const List* useOp
        int location

    ctypedef struct WindowDef:
        NodeTag type
        const char* name
        const char* refname
        const List* partitionClause
        const List* orderClause
        int frameOptions
        const Node* startOffset
        const Node* endOffset
        int location

    ctypedef struct RangeSubselect:
        NodeTag type
        bool lateral
        const Node* subquery
        const Alias* alias

    ctypedef struct RangeFunction:
        NodeTag type
        bool lateral
        bool ordinality
        bool is_rowsfrom
        const List* functions
        const Alias* alias
        const List* coldeflist

    ctypedef struct RangeTableFunc:
        NodeTag type
        bool lateral
        const Node* docexpr
        const Node* rowexpr
        const List* namespaces
        const List* columns
        const Alias* alias
        int location

    ctypedef struct RangeTableFuncCol:
        NodeTag type
        const char* colname
        const TypeName* typeName
        bool for_ordinality
        bool is_not_null
        const Node* colexpr
        const Node* coldefexpr
        int location

    ctypedef struct RangeTableSample:
        NodeTag type
        const Node* relation
        const List* method
        const List* args
        const Node* repeatable
        int location

    ctypedef struct ColumnDef:
        NodeTag type
        const char* colname
        const TypeName* typeName
        const char* compression
        int inhcount
        bool is_local
        bool is_not_null
        bool is_from_type
        char storage
        const Node* raw_default
        const Node* cooked_default
        char identity
        const RangeVar* identitySequence
        char generated
        const CollateClause* collClause
        const List* constraints
        const List* fdwoptions
        int location

    ctypedef struct TableLikeClause:
        NodeTag type
        const RangeVar* relation
        int32_t options

    ctypedef struct IndexElem:
        NodeTag type
        const char* name
        const Node* expr
        const char* indexcolname
        const List* collation
        const List* opclass
        const List* opclassopts
        SortByDir ordering
        SortByNulls nulls_ordering

    ctypedef struct DefElem:
        NodeTag type
        const char* defnamespace
        const char* defname
        const Node* arg
        DefElemAction defaction
        int location

    ctypedef struct LockingClause:
        NodeTag type
        const List* lockedRels
        LockClauseStrength strength
        LockWaitPolicy waitPolicy

    ctypedef struct XmlSerialize:
        NodeTag type
        XmlOptionType xmloption
        const Node* expr
        const TypeName* typeName
        int location

    ctypedef struct PartitionElem:
        NodeTag type
        const char* name
        const Node* expr
        const List* collation
        const List* opclass
        int location

    ctypedef struct PartitionSpec:
        NodeTag type
        const char* strategy
        const List* partParams
        int location

    ctypedef struct PartitionBoundSpec:
        NodeTag type
        char strategy
        bool is_default
        int modulus
        int remainder
        const List* listdatums
        const List* lowerdatums
        const List* upperdatums
        int location

    ctypedef struct PartitionRangeDatum:
        NodeTag type
        PartitionRangeDatumKind kind
        const Node* value
        int location

    ctypedef struct PartitionCmd:
        NodeTag type
        const RangeVar* name
        const PartitionBoundSpec* bound
        bool concurrent

    ctypedef struct RangeTblEntry:
        NodeTag type
        RTEKind rtekind
        char relkind
        int rellockmode
        const TableSampleClause* tablesample
        const Query* subquery
        bool security_barrier
        JoinType jointype
        int joinmergedcols
        const List* joinaliasvars
        const List* joinleftcols
        const List* joinrightcols
        const Alias* join_using_alias
        const List* functions
        bool funcordinality
        const TableFunc* tablefunc
        const List* values_lists
        const char* ctename
        unsigned int ctelevelsup
        bool self_reference
        const List* coltypes
        const List* coltypmods
        const List* colcollations
        const char* enrname
        double enrtuples
        const Alias* alias
        const Alias* eref
        bool lateral
        bool inh
        bool inFromCl
        unsigned int requiredPerms
        const Bitmapset* selectedCols
        const Bitmapset* insertedCols
        const Bitmapset* updatedCols
        const Bitmapset* extraUpdatedCols
        const List* securityQuals

    ctypedef struct RangeTblFunction:
        NodeTag type
        const Node* funcexpr
        int funccolcount
        const List* funccolnames
        const List* funccoltypes
        const List* funccoltypmods
        const List* funccolcollations
        const Bitmapset* funcparams

    ctypedef struct TableSampleClause:
        NodeTag type
        const List* args
        const Expr* repeatable

    ctypedef struct WithCheckOption:
        NodeTag type
        WCOKind kind
        const char* relname
        const char* polname
        const Node* qual
        bool cascaded

    ctypedef struct SortGroupClause:
        NodeTag type
        unsigned int tleSortGroupRef
        bool nulls_first
        bool hashable

    ctypedef struct GroupingSet:
        NodeTag type
        GroupingSetKind kind
        const List* content
        int location

    ctypedef struct WindowClause:
        NodeTag type
        const char* name
        const char* refname
        const List* partitionClause
        const List* orderClause
        int frameOptions
        const Node* startOffset
        const Node* endOffset
        const List* runCondition
        bool inRangeAsc
        bool inRangeNullsFirst
        unsigned int winref
        bool copiedOrder

    ctypedef struct RowMarkClause:
        NodeTag type
        unsigned int rti
        LockClauseStrength strength
        LockWaitPolicy waitPolicy
        bool pushedDown

    ctypedef struct WithClause:
        NodeTag type
        const List* ctes
        bool recursive
        int location

    ctypedef struct InferClause:
        NodeTag type
        const List* indexElems
        const Node* whereClause
        const char* conname
        int location

    ctypedef struct OnConflictClause:
        NodeTag type
        OnConflictAction action
        const InferClause* infer
        const List* targetList
        const Node* whereClause
        int location

    ctypedef struct CTESearchClause:
        NodeTag type
        const List* search_col_list
        bool search_breadth_first
        const char* search_seq_column
        int location

    ctypedef struct CTECycleClause:
        NodeTag type
        const List* cycle_col_list
        const char* cycle_mark_column
        const Node* cycle_mark_value
        const Node* cycle_mark_default
        const char* cycle_path_column
        int location
        int cycle_mark_typmod

    ctypedef struct CommonTableExpr:
        NodeTag type
        const char* ctename
        const List* aliascolnames
        CTEMaterialize ctematerialized
        const Node* ctequery
        const CTESearchClause* search_clause
        const CTECycleClause* cycle_clause
        int location
        bool cterecursive
        int cterefcount
        const List* ctecolnames
        const List* ctecoltypes
        const List* ctecoltypmods
        const List* ctecolcollations

    ctypedef struct MergeWhenClause:
        NodeTag type
        bool matched
        CmdType commandType
        OverridingKind override
        const Node* condition
        const List* targetList
        const List* values

    ctypedef struct MergeAction:
        NodeTag type
        bool matched
        CmdType commandType
        OverridingKind override
        const Node* qual
        const List* targetList
        const List* updateColnos

    ctypedef struct TriggerTransition:
        NodeTag type
        const char* name
        bool isNew
        bool isTable

    ctypedef struct RawStmt:
        NodeTag type
        const Node* stmt
        int stmt_location
        int stmt_len

    ctypedef struct InsertStmt:
        NodeTag type
        const RangeVar* relation
        const List* cols
        const Node* selectStmt
        const OnConflictClause* onConflictClause
        const List* returningList
        const WithClause* withClause
        OverridingKind override

    ctypedef struct DeleteStmt:
        NodeTag type
        const RangeVar* relation
        const List* usingClause
        const Node* whereClause
        const List* returningList
        const WithClause* withClause

    ctypedef struct UpdateStmt:
        NodeTag type
        const RangeVar* relation
        const List* targetList
        const Node* whereClause
        const List* fromClause
        const List* returningList
        const WithClause* withClause

    ctypedef struct MergeStmt:
        NodeTag type
        const RangeVar* relation
        const Node* sourceRelation
        const Node* joinCondition
        const List* mergeWhenClauses
        const WithClause* withClause

    ctypedef struct SelectStmt:
        NodeTag type
        const List* distinctClause
        const IntoClause* intoClause
        const List* targetList
        const List* fromClause
        const Node* whereClause
        const List* groupClause
        bool groupDistinct
        const Node* havingClause
        const List* windowClause
        const List* valuesLists
        const List* sortClause
        const Node* limitOffset
        const Node* limitCount
        LimitOption limitOption
        const List* lockingClause
        const WithClause* withClause
        SetOperation op
        bool all
        const SelectStmt* larg
        const SelectStmt* rarg

    ctypedef struct SetOperationStmt:
        NodeTag type
        SetOperation op
        bool all
        const Node* larg
        const Node* rarg
        const List* colTypes
        const List* colTypmods
        const List* colCollations
        const List* groupClauses

    ctypedef struct ReturnStmt:
        NodeTag type
        const Node* returnval

    ctypedef struct PLAssignStmt:
        NodeTag type
        const char* name
        const List* indirection
        int nnames
        const SelectStmt* val
        int location

    ctypedef struct CreateSchemaStmt:
        NodeTag type
        const char* schemaname
        const RoleSpec* authrole
        const List* schemaElts
        bool if_not_exists

    ctypedef struct AlterTableStmt:
        NodeTag type
        const RangeVar* relation
        const List* cmds
        ObjectType objtype
        bool missing_ok

    ctypedef struct ReplicaIdentityStmt:
        NodeTag type
        char identity_type
        const char* name

    ctypedef struct AlterTableCmd:
        NodeTag type
        AlterTableType subtype
        const char* name
        int16_t num
        const RoleSpec* newowner
        const Node* def_ "def"
        DropBehavior behavior
        bool missing_ok
        bool recurse

    ctypedef struct AlterCollationStmt:
        NodeTag type
        const List* collname

    ctypedef struct AlterDomainStmt:
        NodeTag type
        char subtype
        const List* typeName
        const char* name
        const Node* def_ "def"
        DropBehavior behavior
        bool missing_ok

    ctypedef struct GrantStmt:
        NodeTag type
        bool is_grant
        GrantTargetType targtype
        ObjectType objtype
        const List* objects
        const List* privileges
        const List* grantees
        bool grant_option
        const RoleSpec* grantor
        DropBehavior behavior

    ctypedef struct ObjectWithArgs:
        NodeTag type
        const List* objname
        const List* objargs
        const List* objfuncargs
        bool args_unspecified

    ctypedef struct AccessPriv:
        NodeTag type
        const char* priv_name
        const List* cols

    ctypedef struct GrantRoleStmt:
        NodeTag type
        const List* granted_roles
        const List* grantee_roles
        bool is_grant
        bool admin_opt
        const RoleSpec* grantor
        DropBehavior behavior

    ctypedef struct AlterDefaultPrivilegesStmt:
        NodeTag type
        const List* options
        const GrantStmt* action

    ctypedef struct CopyStmt:
        NodeTag type
        const RangeVar* relation
        const Node* query
        const List* attlist
        bool is_from
        bool is_program
        const char* filename
        const List* options
        const Node* whereClause

    ctypedef struct VariableSetStmt:
        NodeTag type
        VariableSetKind kind
        const char* name
        const List* args
        bool is_local

    ctypedef struct VariableShowStmt:
        NodeTag type
        const char* name

    ctypedef struct CreateStmt:
        NodeTag type
        const RangeVar* relation
        const List* tableElts
        const List* inhRelations
        const PartitionBoundSpec* partbound
        const PartitionSpec* partspec
        const TypeName* ofTypename
        const List* constraints
        const List* options
        OnCommitAction oncommit
        const char* tablespacename
        const char* accessMethod
        bool if_not_exists

    ctypedef struct Constraint:
        NodeTag type
        ConstrType contype
        const char* conname
        bool deferrable
        bool initdeferred
        int location
        bool is_no_inherit
        const Node* raw_expr
        const char* cooked_expr
        char generated_when
        bool nulls_not_distinct
        const List* keys
        const List* including
        const List* exclusions
        const List* options
        const char* indexname
        const char* indexspace
        bool reset_default_tblspc
        const char* access_method
        const Node* where_clause
        const RangeVar* pktable
        const List* fk_attrs
        const List* pk_attrs
        char fk_matchtype
        char fk_upd_action
        char fk_del_action
        const List* fk_del_set_cols
        const List* old_conpfeqop
        bool skip_validation
        bool initially_valid

    ctypedef struct CreateTableSpaceStmt:
        NodeTag type
        const char* tablespacename
        const RoleSpec* owner
        const char* location
        const List* options

    ctypedef struct DropTableSpaceStmt:
        NodeTag type
        const char* tablespacename
        bool missing_ok

    ctypedef struct AlterTableSpaceOptionsStmt:
        NodeTag type
        const char* tablespacename
        const List* options
        bool isReset

    ctypedef struct AlterTableMoveAllStmt:
        NodeTag type
        const char* orig_tablespacename
        ObjectType objtype
        const List* roles
        const char* new_tablespacename
        bool nowait

    ctypedef struct CreateExtensionStmt:
        NodeTag type
        const char* extname
        bool if_not_exists
        const List* options

    ctypedef struct AlterExtensionStmt:
        NodeTag type
        const char* extname
        const List* options

    ctypedef struct AlterExtensionContentsStmt:
        NodeTag type
        const char* extname
        int action
        ObjectType objtype
        const Node* object

    ctypedef struct CreateFdwStmt:
        NodeTag type
        const char* fdwname
        const List* func_options
        const List* options

    ctypedef struct AlterFdwStmt:
        NodeTag type
        const char* fdwname
        const List* func_options
        const List* options

    ctypedef struct CreateForeignServerStmt:
        NodeTag type
        const char* servername
        const char* servertype
        const char* version
        const char* fdwname
        bool if_not_exists
        const List* options

    ctypedef struct AlterForeignServerStmt:
        NodeTag type
        const char* servername
        const char* version
        const List* options
        bool has_version

    ctypedef struct CreateForeignTableStmt:
        CreateStmt base
        const char* servername
        const List* options

    ctypedef struct CreateUserMappingStmt:
        NodeTag type
        const RoleSpec* user
        const char* servername
        bool if_not_exists
        const List* options

    ctypedef struct AlterUserMappingStmt:
        NodeTag type
        const RoleSpec* user
        const char* servername
        const List* options

    ctypedef struct DropUserMappingStmt:
        NodeTag type
        const RoleSpec* user
        const char* servername
        bool missing_ok

    ctypedef struct ImportForeignSchemaStmt:
        NodeTag type
        const char* server_name
        const char* remote_schema
        const char* local_schema
        ImportForeignSchemaType list_type
        const List* table_list
        const List* options

    ctypedef struct CreatePolicyStmt:
        NodeTag type
        const char* policy_name
        const RangeVar* table
        const char* cmd_name
        bool permissive
        const List* roles
        const Node* qual
        const Node* with_check

    ctypedef struct AlterPolicyStmt:
        NodeTag type
        const char* policy_name
        const RangeVar* table
        const List* roles
        const Node* qual
        const Node* with_check

    ctypedef struct CreateAmStmt:
        NodeTag type
        const char* amname
        const List* handler_name
        char amtype

    ctypedef struct CreateTrigStmt:
        NodeTag type
        bool replace
        bool isconstraint
        const char* trigname
        const RangeVar* relation
        const List* funcname
        const List* args
        bool row
        int16_t timing
        int16_t events
        const List* columns
        const Node* whenClause
        const List* transitionRels
        bool deferrable
        bool initdeferred
        const RangeVar* constrrel

    ctypedef struct CreateEventTrigStmt:
        NodeTag type
        const char* trigname
        const char* eventname
        const List* whenclause
        const List* funcname

    ctypedef struct AlterEventTrigStmt:
        NodeTag type
        const char* trigname
        char tgenabled

    ctypedef struct CreatePLangStmt:
        NodeTag type
        bool replace
        const char* plname
        const List* plhandler
        const List* plinline
        const List* plvalidator
        bool pltrusted

    ctypedef struct CreateRoleStmt:
        NodeTag type
        RoleStmtType stmt_type
        const char* role
        const List* options

    ctypedef struct AlterRoleStmt:
        NodeTag type
        const RoleSpec* role
        const List* options
        int action

    ctypedef struct AlterRoleSetStmt:
        NodeTag type
        const RoleSpec* role
        const char* database
        const VariableSetStmt* setstmt

    ctypedef struct DropRoleStmt:
        NodeTag type
        const List* roles
        bool missing_ok

    ctypedef struct CreateSeqStmt:
        NodeTag type
        const RangeVar* sequence
        const List* options
        bool for_identity
        bool if_not_exists

    ctypedef struct AlterSeqStmt:
        NodeTag type
        const RangeVar* sequence
        const List* options
        bool for_identity
        bool missing_ok

    ctypedef struct DefineStmt:
        NodeTag type
        ObjectType kind
        bool oldstyle
        const List* defnames
        const List* args
        const List* definition
        bool if_not_exists
        bool replace

    ctypedef struct CreateDomainStmt:
        NodeTag type
        const List* domainname
        const TypeName* typeName
        const CollateClause* collClause
        const List* constraints

    ctypedef struct CreateOpClassStmt:
        NodeTag type
        const List* opclassname
        const List* opfamilyname
        const char* amname
        const TypeName* datatype
        const List* items
        bool isDefault

    ctypedef struct CreateOpClassItem:
        NodeTag type
        int itemtype
        const ObjectWithArgs* name
        int number
        const List* order_family
        const List* class_args
        const TypeName* storedtype

    ctypedef struct CreateOpFamilyStmt:
        NodeTag type
        const List* opfamilyname
        const char* amname

    ctypedef struct AlterOpFamilyStmt:
        NodeTag type
        const List* opfamilyname
        const char* amname
        bool isDrop
        const List* items

    ctypedef struct DropStmt:
        NodeTag type
        const List* objects
        ObjectType removeType
        DropBehavior behavior
        bool missing_ok
        bool concurrent

    ctypedef struct TruncateStmt:
        NodeTag type
        const List* relations
        bool restart_seqs
        DropBehavior behavior

    ctypedef struct CommentStmt:
        NodeTag type
        ObjectType objtype
        const Node* object
        const char* comment

    ctypedef struct SecLabelStmt:
        NodeTag type
        ObjectType objtype
        const Node* object
        const char* provider
        const char* label

    ctypedef struct DeclareCursorStmt:
        NodeTag type
        const char* portalname
        int options
        const Node* query

    ctypedef struct ClosePortalStmt:
        NodeTag type
        const char* portalname

    ctypedef struct FetchStmt:
        NodeTag type
        FetchDirection direction
        long howMany
        const char* portalname
        bool ismove

    ctypedef struct IndexStmt:
        NodeTag type
        const char* idxname
        const RangeVar* relation
        const char* accessMethod
        const char* tableSpace
        const List* indexParams
        const List* indexIncludingParams
        const List* options
        const Node* whereClause
        const List* excludeOpNames
        const char* idxcomment
        unsigned int oldCreateSubid
        unsigned int oldFirstRelfilenodeSubid
        bool unique
        bool nulls_not_distinct
        bool primary
        bool isconstraint
        bool deferrable
        bool initdeferred
        bool transformed
        bool concurrent
        bool if_not_exists
        bool reset_default_tblspc

    ctypedef struct CreateStatsStmt:
        NodeTag type
        const List* defnames
        const List* stat_types
        const List* exprs
        const List* relations
        const char* stxcomment
        bool transformed
        bool if_not_exists

    ctypedef struct StatsElem:
        NodeTag type
        const char* name
        const Node* expr

    ctypedef struct AlterStatsStmt:
        NodeTag type
        const List* defnames
        int stxstattarget
        bool missing_ok

    ctypedef struct CreateFunctionStmt:
        NodeTag type
        bool is_procedure
        bool replace
        const List* funcname
        const List* parameters
        const TypeName* returnType
        const List* options
        const Node* sql_body

    ctypedef struct FunctionParameter:
        NodeTag type
        const char* name
        const TypeName* argType
        FunctionParameterMode mode
        const Node* defexpr

    ctypedef struct AlterFunctionStmt:
        NodeTag type
        ObjectType objtype
        const ObjectWithArgs* func
        const List* actions

    ctypedef struct DoStmt:
        NodeTag type
        const List* args

    ctypedef struct InlineCodeBlock:
        NodeTag type
        const char* source_text
        bool langIsTrusted
        bool atomic

    ctypedef struct CallStmt:
        NodeTag type
        const FuncCall* funccall
        const FuncExpr* funcexpr
        const List* outargs

    ctypedef struct CallContext:
        NodeTag type
        bool atomic

    ctypedef struct RenameStmt:
        NodeTag type
        ObjectType renameType
        ObjectType relationType
        const RangeVar* relation
        const Node* object
        const char* subname
        const char* newname
        DropBehavior behavior
        bool missing_ok

    ctypedef struct AlterObjectDependsStmt:
        NodeTag type
        ObjectType objectType
        const RangeVar* relation
        const Node* object
        const String* extname
        bool remove

    ctypedef struct AlterObjectSchemaStmt:
        NodeTag type
        ObjectType objectType
        const RangeVar* relation
        const Node* object
        const char* newschema
        bool missing_ok

    ctypedef struct AlterOwnerStmt:
        NodeTag type
        ObjectType objectType
        const RangeVar* relation
        const Node* object
        const RoleSpec* newowner

    ctypedef struct AlterOperatorStmt:
        NodeTag type
        const ObjectWithArgs* opername
        const List* options

    ctypedef struct AlterTypeStmt:
        NodeTag type
        const List* typeName
        const List* options

    ctypedef struct RuleStmt:
        NodeTag type
        const RangeVar* relation
        const char* rulename
        const Node* whereClause
        CmdType event
        bool instead
        const List* actions
        bool replace

    ctypedef struct NotifyStmt:
        NodeTag type
        const char* conditionname
        const char* payload

    ctypedef struct ListenStmt:
        NodeTag type
        const char* conditionname

    ctypedef struct UnlistenStmt:
        NodeTag type
        const char* conditionname

    ctypedef struct TransactionStmt:
        NodeTag type
        TransactionStmtKind kind
        const List* options
        const char* savepoint_name
        const char* gid
        bool chain

    ctypedef struct CompositeTypeStmt:
        NodeTag type
        const RangeVar* typevar
        const List* coldeflist

    ctypedef struct CreateEnumStmt:
        NodeTag type
        const List* typeName
        const List* vals

    ctypedef struct CreateRangeStmt:
        NodeTag type
        const List* typeName
        const List* params

    ctypedef struct AlterEnumStmt:
        NodeTag type
        const List* typeName
        const char* oldVal
        const char* newVal
        const char* newValNeighbor
        bool newValIsAfter
        bool skipIfNewValExists

    ctypedef struct ViewStmt:
        NodeTag type
        const RangeVar* view
        const List* aliases
        const Node* query
        bool replace
        const List* options
        ViewCheckOption withCheckOption

    ctypedef struct LoadStmt:
        NodeTag type
        const char* filename

    ctypedef struct CreatedbStmt:
        NodeTag type
        const char* dbname
        const List* options

    ctypedef struct AlterDatabaseStmt:
        NodeTag type
        const char* dbname
        const List* options

    ctypedef struct AlterDatabaseRefreshCollStmt:
        NodeTag type
        const char* dbname

    ctypedef struct AlterDatabaseSetStmt:
        NodeTag type
        const char* dbname
        const VariableSetStmt* setstmt

    ctypedef struct DropdbStmt:
        NodeTag type
        const char* dbname
        bool missing_ok
        const List* options

    ctypedef struct AlterSystemStmt:
        NodeTag type
        const VariableSetStmt* setstmt

    ctypedef struct ClusterStmt:
        NodeTag type
        const RangeVar* relation
        const char* indexname
        const List* params

    ctypedef struct VacuumStmt:
        NodeTag type
        const List* options
        const List* rels
        bool is_vacuumcmd

    ctypedef struct VacuumRelation:
        NodeTag type
        const RangeVar* relation
        const List* va_cols

    ctypedef struct ExplainStmt:
        NodeTag type
        const Node* query
        const List* options

    ctypedef struct CreateTableAsStmt:
        NodeTag type
        const Node* query
        const IntoClause* into
        ObjectType objtype
        bool is_select_into
        bool if_not_exists

    ctypedef struct RefreshMatViewStmt:
        NodeTag type
        bool concurrent
        bool skipData
        const RangeVar* relation

    ctypedef struct CheckPointStmt:
        NodeTag type

    ctypedef struct DiscardStmt:
        NodeTag type
        DiscardMode target

    ctypedef struct LockStmt:
        NodeTag type
        const List* relations
        int mode
        bool nowait

    ctypedef struct ConstraintsSetStmt:
        NodeTag type
        const List* constraints
        bool deferred

    ctypedef struct ReindexStmt:
        NodeTag type
        ReindexObjectType kind
        const RangeVar* relation
        const char* name
        const List* params

    ctypedef struct CreateConversionStmt:
        NodeTag type
        const List* conversion_name
        const char* for_encoding_name
        const char* to_encoding_name
        const List* func_name
        bool def_ "def"

    ctypedef struct CreateCastStmt:
        NodeTag type
        const TypeName* sourcetype
        const TypeName* targettype
        const ObjectWithArgs* func
        CoercionContext context
        bool inout

    ctypedef struct CreateTransformStmt:
        NodeTag type
        bool replace
        const TypeName* type_name
        const char* lang
        const ObjectWithArgs* fromsql
        const ObjectWithArgs* tosql

    ctypedef struct PrepareStmt:
        NodeTag type
        const char* name
        const List* argtypes
        const Node* query

    ctypedef struct ExecuteStmt:
        NodeTag type
        const char* name
        const List* params

    ctypedef struct DeallocateStmt:
        NodeTag type
        const char* name

    ctypedef struct DropOwnedStmt:
        NodeTag type
        const List* roles
        DropBehavior behavior

    ctypedef struct ReassignOwnedStmt:
        NodeTag type
        const List* roles
        const RoleSpec* newrole

    ctypedef struct AlterTSDictionaryStmt:
        NodeTag type
        const List* dictname
        const List* options

    ctypedef struct AlterTSConfigurationStmt:
        NodeTag type
        AlterTSConfigType kind
        const List* cfgname
        const List* tokentype
        const List* dicts
        bool override
        bool replace
        bool missing_ok

    ctypedef struct PublicationTable:
        NodeTag type
        const RangeVar* relation
        const Node* whereClause
        const List* columns

    ctypedef struct PublicationObjSpec:
        NodeTag type
        PublicationObjSpecType pubobjtype
        const char* name
        const PublicationTable* pubtable
        int location

    ctypedef struct CreatePublicationStmt:
        NodeTag type
        const char* pubname
        const List* options
        const List* pubobjects
        bool for_all_tables

    ctypedef struct AlterPublicationStmt:
        NodeTag type
        const char* pubname
        const List* options
        const List* pubobjects
        bool for_all_tables
        AlterPublicationAction action

    ctypedef struct CreateSubscriptionStmt:
        NodeTag type
        const char* subname
        const char* conninfo
        const List* publication
        const List* options

    ctypedef struct AlterSubscriptionStmt:
        NodeTag type
        AlterSubscriptionType kind
        const char* subname
        const char* conninfo
        const List* publication
        const List* options

    ctypedef struct DropSubscriptionStmt:
        NodeTag type
        const char* subname
        bool missing_ok
        DropBehavior behavior


cdef extern from "nodes/primnodes.h":
    ctypedef struct Alias:
        NodeTag type
        const char* aliasname
        const List* colnames

    ctypedef struct RangeVar:
        NodeTag type
        const char* catalogname
        const char* schemaname
        const char* relname
        bool inh
        char relpersistence
        const Alias* alias
        int location

    ctypedef struct TableFunc:
        NodeTag type
        const List* ns_uris
        const List* ns_names
        const Node* docexpr
        const Node* rowexpr
        const List* colnames
        const List* coltypes
        const List* coltypmods
        const List* colcollations
        const List* colexprs
        const List* coldefexprs
        const Bitmapset* notnulls
        int ordinalitycol
        int location

    ctypedef struct IntoClause:
        NodeTag type
        const RangeVar* rel
        const List* colNames
        const char* accessMethod
        const List* options
        OnCommitAction onCommit
        const char* tableSpaceName
        const Node* viewQuery
        bool skipData

    ctypedef struct Expr:
        NodeTag type

    ctypedef struct Var:
        int varno
        int varattno
        int32_t vartypmod
        unsigned int varlevelsup
        unsigned int varnosyn
        int varattnosyn
        int location

    ctypedef struct Param:
        ParamKind paramkind
        int paramid
        int32_t paramtypmod
        int location

    ctypedef struct Aggref:
        const List* aggargtypes
        const List* aggdirectargs
        const List* args
        const List* aggorder
        const List* aggdistinct
        const Expr* aggfilter
        bool aggstar
        bool aggvariadic
        char aggkind
        unsigned int agglevelsup
        AggSplit aggsplit
        int aggno
        int aggtransno
        int location

    ctypedef struct GroupingFunc:
        const List* args
        const List* refs
        const List* cols
        unsigned int agglevelsup
        int location

    ctypedef struct WindowFunc:
        const List* args
        const Expr* aggfilter
        unsigned int winref
        bool winstar
        bool winagg
        int location

    ctypedef struct SubscriptingRef:
        int32_t reftypmod
        const List* refupperindexpr
        const List* reflowerindexpr
        const Expr* refexpr
        const Expr* refassgnexpr

    ctypedef struct FuncExpr:
        bool funcretset
        bool funcvariadic
        CoercionForm funcformat
        const List* args
        int location

    ctypedef struct NamedArgExpr:
        const Expr* arg
        const char* name
        int argnumber
        int location

    ctypedef struct OpExpr:
        bool opretset
        const List* args
        int location

    ctypedef struct ScalarArrayOpExpr:
        bool useOr
        const List* args
        int location

    ctypedef struct BoolExpr:
        BoolExprType boolop
        const List* args
        int location

    ctypedef struct SubLink:
        SubLinkType subLinkType
        int subLinkId
        const Node* testexpr
        const List* operName
        const Node* subselect
        int location

    ctypedef struct SubPlan:
        SubLinkType subLinkType
        const Node* testexpr
        const List* paramIds
        int plan_id
        const char* plan_name
        int32_t firstColTypmod
        bool useHashTable
        bool unknownEqFalse
        bool parallel_safe
        const List* setParam
        const List* parParam
        const List* args
        double startup_cost
        double per_call_cost

    ctypedef struct AlternativeSubPlan:
        const List* subplans

    ctypedef struct FieldSelect:
        const Expr* arg
        int fieldnum
        int32_t resulttypmod

    ctypedef struct FieldStore:
        const Expr* arg
        const List* newvals
        const List* fieldnums

    ctypedef struct RelabelType:
        const Expr* arg
        int32_t resulttypmod
        CoercionForm relabelformat
        int location

    ctypedef struct CoerceViaIO:
        const Expr* arg
        CoercionForm coerceformat
        int location

    ctypedef struct ArrayCoerceExpr:
        const Expr* arg
        const Expr* elemexpr
        int32_t resulttypmod
        CoercionForm coerceformat
        int location

    ctypedef struct ConvertRowtypeExpr:
        const Expr* arg
        CoercionForm convertformat
        int location

    ctypedef struct CollateExpr:
        const Expr* arg
        int location

    ctypedef struct CaseExpr:
        const Expr* arg
        const List* args
        const Expr* defresult
        int location

    ctypedef struct CaseWhen:
        const Expr* expr
        const Expr* result
        int location

    ctypedef struct CaseTestExpr:
        int32_t typeMod

    ctypedef struct ArrayExpr:
        const List* elements
        bool multidims
        int location

    ctypedef struct RowExpr:
        const List* args
        CoercionForm row_format
        const List* colnames
        int location

    ctypedef struct RowCompareExpr:
        RowCompareType rctype
        const List* opnos
        const List* opfamilies
        const List* inputcollids
        const List* largs
        const List* rargs

    ctypedef struct CoalesceExpr:
        const List* args
        int location

    ctypedef struct MinMaxExpr:
        MinMaxOp op
        const List* args
        int location

    ctypedef struct SQLValueFunction:
        SQLValueFunctionOp op
        int32_t typmod
        int location

    ctypedef struct XmlExpr:
        XmlExprOp op
        const char* name
        const List* named_args
        const List* arg_names
        const List* args
        XmlOptionType xmloption
        int32_t typmod
        int location

    ctypedef struct NullTest:
        const Expr* arg
        NullTestType nulltesttype
        bool argisrow
        int location

    ctypedef struct BooleanTest:
        const Expr* arg
        BoolTestType booltesttype
        int location

    ctypedef struct CoerceToDomain:
        const Expr* arg
        int32_t resulttypmod
        CoercionForm coercionformat
        int location

    ctypedef struct CoerceToDomainValue:
        int32_t typeMod
        int location

    ctypedef struct SetToDefault:
        int32_t typeMod
        int location

    ctypedef struct CurrentOfExpr:
        unsigned int cvarno
        const char* cursor_name
        int cursor_param

    ctypedef struct InferenceElem:
        const Node* expr

    ctypedef struct TargetEntry:
        const Expr* expr
        int resno
        const char* resname
        unsigned int ressortgroupref
        int resorigcol
        bool resjunk

    ctypedef struct RangeTblRef:
        NodeTag type
        int rtindex

    ctypedef struct JoinExpr:
        NodeTag type
        JoinType jointype
        bool isNatural
        const Node* larg
        const Node* rarg
        const List* usingClause
        const Alias* join_using_alias
        const Node* quals
        const Alias* alias
        int rtindex

    ctypedef struct FromExpr:
        NodeTag type
        const List* fromlist
        const Node* quals

    ctypedef struct OnConflictExpr:
        NodeTag type
        OnConflictAction action
        const List* arbiterElems
        const Node* arbiterWhere
        const List* onConflictSet
        const Node* onConflictWhere
        int exclRelIndex
        const List* exclRelTlist


cdef extern from "nodes/value.h":
    ctypedef struct Integer:
        long ival

    ctypedef struct Float:
        const char* fval

    ctypedef struct Boolean:
        bool boolval

    ctypedef struct String:
        const char* sval

    ctypedef struct BitString:
        const char* bsval

    ctypedef struct A_Const:
        bool isnull
        ValUnion val



# -*- coding: utf-8; mode: python -*-
# :Project:   pglast -- PostgreSQL Languages AST
# :Created:   mer 02 ago 2017 15:20:43 CEST
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2020, 2021, 2022 Lele Gaifax
#

from pathlib import Path
import subprocess

from setuptools import Extension, setup, find_packages
from setuptools.command.build_ext import build_ext
try:
    from Cython.Build import cythonize
except ImportError:
    cythonize = lambda e: e
    extension_source = 'pglast/parser.c'
else:
    extension_source = 'pglast/parser.pyx'


here = Path(__file__).absolute().parent
with (here / 'README.rst').open(encoding='utf-8') as f:
    README = f.read()
with (here / 'CHANGES.rst').open(encoding='utf-8') as f:
    CHANGES = f.read()
with (here / 'version.txt').open(encoding='utf-8') as f:
    VERSION = f.read().strip()


LIBPG_QUERY_DIR = str(Path('libpg_query'))
INCLUDE_DIR = str(Path('libpg_query') / 'src' / 'postgres' / 'include')
VENDOR_DIR = str(Path('libpg_query') / 'vendor')


class BuildLibPgQueryFirst(build_ext):
    def run(self):
        subprocess.check_call(['make', '-s', '-C', LIBPG_QUERY_DIR, 'build'])
        super().run()


setup(
    name="pglast",
    version=VERSION,
    url="https://github.com/lelit/pglast",

    description="PostgreSQL Languages AST and statements prettifier",
    long_description=README + '\n\n' + CHANGES,
    long_description_content_type='text/x-rst',

    author="Lele Gaifax",
    author_email="lele@metapensiero.it",

    license="GPLv3+",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3 :: Only",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: SQL",
        # "Programming Language :: PL/SQL",
        "Topic :: Database",
        "Topic :: Utilities",
        ],
    keywords="postgresql parser sql prettifier",

    packages=find_packages('.'),

    cmdclass={'build_ext': BuildLibPgQueryFirst},
    ext_modules=cythonize([
        Extension('pglast.parser', [extension_source],
                  libraries=['pg_query'],
                  include_dirs=[LIBPG_QUERY_DIR, INCLUDE_DIR, VENDOR_DIR],
                  library_dirs=[LIBPG_QUERY_DIR]),
    ]),

    install_requires=[
        'setuptools',
    ],
    extras_require={
        'dev': [
            'cython',
            'metapensiero.tool.bump_version',
            'pycparser',
            'readme_renderer',
        ]
    },
    entry_points="""\
    [console_scripts]
    pgpp = pglast.__main__:main
    """,
)



# -*- coding: utf-8 -*-
# :Project:   pglast -- Extract printer functions doc
# :Created:   gio 09 nov 2017 12:56:35 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2017, 2018, 2019, 2020, 2022 Lele Gaifax
#

from datetime import date
from os.path import basename, splitext
from re import match
import subprocess


RST_HEADER = """\
.. -*- coding: utf-8 -*-
.. :Project:   pglast -- DO NOT EDIT: generated automatically
.. :Author:    Lele Gaifax <lele@metapensiero.it>
.. :License:   GNU General Public License version 3 or later
.. :Copyright: © 2017-%d Lele Gaifax
..

================================================%%(extra_decoration)s
 :mod:`pglast.printers.%%(mod_name)s` --- %%(mod_nick_name)s printer functions
================================================%%(extra_decoration)s

.. module:: pglast.printers.%%(mod_name)s
   :synopsis: %%(mod_nick_name)s printer functions
""" % date.today().year


def get_libpg_query_info():
    "Return a tuple with (version, baseurl) of the libpg_query library."

    version = subprocess.check_output(['git', 'describe', '--all', '--long'],
                                      cwd='libpg_query')
    version = version.decode('utf-8').strip().split('/')[-1]
    remote = subprocess.check_output(['git', 'remote', 'get-url', 'origin'],
                                     cwd='libpg_query')
    remote = remote.decode('utf-8')
    baseurl = '%s/blob/%s/' % (remote[:-5], version[-7:])
    return version, baseurl


def extract_toc(headers):
    "Extract the nodes position in the headers."

    toc = {}

    for header in headers:
        with open(header, encoding='utf-8') as f:
            content = f.read()

        for lineno, line in enumerate(content.splitlines(), 1):
            if line.startswith('typedef struct '):
                m = match(r'typedef struct\s+([\w_]+)', line)
                if m is not None:
                    toc[m.group(1)] = (header, lineno)
            elif line.startswith('struct PartitionBoundSpec'):
                # For some reason, this is an exception: unlike all other nodes
                # in nodes/parsenodes.h, this is not typedef'd right away, but
                # rather in partitioning/partdefs.h
                toc['PartitionBoundSpec'] = (header, lineno)

    return toc


def extract_printers(source):
    "Extract printer functions implemented in `source`."

    printers = []

    with open(source, encoding='utf-8') as f:
        content = f.read()

    nps = []
    lines = iter(content.splitlines())
    for line in lines:
        if line.startswith('@node_printer('):
            while not line.endswith(')'):
                line += next(lines)
            m = match(r'@node_printer\((.*,\s)?([\w._]+)\)', line)
            assert m is not None, f"Could not parse @node_printer(): {line}"
            scope = m.group(1)
            if scope:
                scope = scope.lstrip('(').rstrip('), ').split(',')
                scope = tuple(s.strip() for s in scope)
            node = m.group(2)
            nps.append((scope, node))
        elif line.startswith('def '):
            if nps:
                funcname = line[4:].split('(', 1)[0]
                for scope, node in nps:
                    printers.append((scope, node, funcname))
                nps = []

    return printers


def workhorse(args):
    libpg_query_version, libpg_query_baseurl = get_libpg_query_info()
    toc = extract_toc(args.headers)
    printers = extract_printers(args.source)
    with open(args.rstdoc, 'w', encoding='utf-8') as rstdoc:
        source_fname = basename(args.source)
        mod_name = splitext(source_fname)[0]
        rstdoc.write(RST_HEADER % dict(
            mod_name=mod_name, mod_nick_name=mod_name.upper(),
            extra_decoration='='*(len(mod_name) * 2)))
        for scope, node, funcname in printers:
            if scope:
                scoped = ", when it is inside "
                snode = scope[0][4:]
                header, lineno = toc[snode]
                header_url = libpg_query_baseurl + header[12:]
                scoped += "a `%s <%s#L%d>`__" % (snode, header_url, lineno)
                for snode in scope[1:]:
                    snode = snode[4:]
                    header, lineno = toc[snode]
                    header_url = libpg_query_baseurl + header[12:]
                    scoped += " or a `%s <%s#L%d>`__" % (snode, header_url, lineno)
                scoped += ','
            else:
                scoped = ''
            node = node[4:]
            header, lineno = toc[node]
            header_url = libpg_query_baseurl + header[12:]
            if scope:
                for snode in scope:
                    snode = snode[4:]
                    rstdoc.write('\n.. index::\n   pair: %s;%s\n' % (snode, node))
            else:
                rstdoc.write('\n.. index:: %s\n' % node)
            rstdoc.write('\n.. function:: %s(node, output)\n' % funcname)
            rstdoc.write('\n   Pretty print a `node` of type `%s <%s#L%d>`__%s to the `output`'
                         ' stream.\n' % (node, header_url, lineno, scoped))


def main():
    from argparse import ArgumentParser, ArgumentDefaultsHelpFormatter

    parser = ArgumentParser(description="Printer functions doc extractor",
                            formatter_class=ArgumentDefaultsHelpFormatter)

    parser.add_argument('source',
                        help="Python source module to be documented")
    parser.add_argument('rstdoc',
                        help="reST documentation to be created")
    parser.add_argument('headers', nargs='+',
                        help="The PG headers containing nodes definitions")
    args = parser.parse_args()

    workhorse(args)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   pglast -- Assert printers eat their own result
# :Created:   dom 17 mar 2019 09:24:11 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2019, 2021, 2022 Lele Gaifax
#

from pathlib import Path
from re import sub, subn

import pytest

from pglast import parse_sql, split
from pglast.parser import ParseError
from pglast.stream import RawStream, IndentedStream
import pglast.printers  # noqa


this = Path(__file__)
this_dir = this.parent
tests_dir = this_dir / this.stem


def statements(src):
    lineno = 1
    for statement in src.read_text().split('\n\n'):
        yield lineno, statement.strip()
        lineno += statement.count('\n') + 2


def make_id(arg):
    if isinstance(arg, Path):
        return str(arg.relative_to(this_dir))
    elif isinstance(arg, int):
        return str(arg)


@pytest.mark.parametrize('src,lineno,statement',
                         ((src, lineno, statement)
                          for src in sorted(tests_dir.glob('**/*.sql'))
                          for (lineno, statement) in statements(src)),
                         ids=make_id)
def test_printers_roundtrip(src, lineno, statement):
    try:
        orig_ast = parse_sql(statement)
    except:  # noqa
        raise RuntimeError("%s:%d:Could not parse %r" % (src, lineno, statement))

    serialized = RawStream()(orig_ast)
    try:
        serialized_ast = parse_sql(serialized)
    except:  # noqa
        raise RuntimeError("%s:%d:Could not reparse %r" % (src, lineno, serialized))

    assert orig_ast == serialized_ast, "%s:%s:%r != %r" % (src, lineno, statement, serialized)

    indented = IndentedStream()(orig_ast)
    try:
        indented_ast = parse_sql(indented)
    except:  # noqa
        raise RuntimeError("%s:%d:Could not reparse %r" % (src, lineno, indented))

    assert orig_ast == indented_ast, "%s:%d:%r != %r" % (src, lineno, statement, indented)

    # Run ``pytest -s tests/`` to see the following output
    print()
    print(indented)


@pytest.mark.parametrize('src,lineno,statement',
                         ((src, lineno, statement)
                          for src in sorted(tests_dir.glob('**/*.sql'))
                          for (lineno, statement) in statements(src)),
                         ids=make_id)
def test_stream_call_with_single_node(src, lineno, statement):
    # See https://github.com/lelit/pglast/pull/10 and https://github.com/lelit/pglast/issues/79
    try:
        parsed = parse_sql(statement)
    except:  # noqa
        raise RuntimeError("%s:%d:Could not parse %r" % (src, lineno, statement))
    for rawstmt in parsed:
        stmt = rawstmt.stmt
        try:
            RawStream()(stmt)
        except Exception:
            raise AssertionError('Could not serialize single statement %r' % stmt)


pg_regressions_dir = this_dir / '..' / 'libpg_query' / 'test' / 'sql' / 'postgres_regress'

# Following scripts contain intentional errors, and it's difficult to isolate them
skip_due_syntax_error = {'unicode.sql'}


@pytest.mark.parametrize('filename',
                         (src.name for src in sorted(pg_regressions_dir.glob('*.sql'))
                          if src.name not in skip_due_syntax_error),
                         ids=make_id)
def test_pg_regress_corpus(filename):
    # we do this dance to minimize the length of the test name
    src = pg_regressions_dir / filename
    source = src.read_text()
    try:
        slices = split(source, with_parser=False, only_slices=True)
    except Exception:
        return

    for slice in slices:
        stmt = source[slice]
        lineno = source[:slice.start].count('\n') + 1

        # Remove comments and replace \n to make it easier cut&pasting the stmt
        # in `pgpp -t`

        trimmed_stmt = stmt.strip()
        while trimmed_stmt.startswith('--'):
            trimmed_stmt = sub(r'--.*\n', '', trimmed_stmt)
            lineno += 1
        trimmed_stmt = subn(r'[\n\t]+', ' ', trimmed_stmt)[0].strip()

        rel_src = src.relative_to(this_dir / '..')
        try:
            orig_ast = parse_sql(stmt)
        except ParseError:
            continue
        except Exception as e:
            raise RuntimeError("Statement “%s” from %s at line %d, could not parse: %s"
                               % (trimmed_stmt, rel_src, lineno, e))

        try:
            serialized = RawStream()(orig_ast)
        except NotImplementedError as e:
            raise NotImplementedError("Statement “%s” from %s at line %d, could not reprint: %s"
                                      % (trimmed_stmt, rel_src, lineno, e))
        except Exception as e:
            raise RuntimeError("Statement “%s” from %s at line %d, could not reprint: %s"
                               % (trimmed_stmt, rel_src, lineno, e))

        try:
            serialized_ast = parse_sql(serialized)
        except Exception as e:
            raise RuntimeError("Statement “%s” from %s at line %d, could not reparse %r: %s"
                               % (trimmed_stmt, rel_src, lineno, serialized, e))

        assert orig_ast == serialized_ast, "Statement “%s” from %s at line %d != %r" % (
            trimmed_stmt, rel_src, lineno, serialized)
